# -*- coding: utf-8 -*-
# Copyright (C) 2011  Alibaba Cloud Computing
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.
#
import mmap
import os

from api import APIProxy
from job import Job
from uploader import Uploader
from utils import *


class Vault(object):
    _MEGABYTE = 1024 * 1024

    NormalUploadThreshold = 64 * _MEGABYTE

    ResponseDataParser = (('CreationDate', 'creation_date', None),
                          ('LastInventoryDate', 'last_inventory_date', None),
                          ('NumberOfArchives', 'number_of_archives', 0),
                          ('SizeInBytes', 'size', 0),
                          ('VaultId', 'id', None),
                          ('VaultName', 'name', None))

    def __init__(self, api, response):
        self.api = api
        for response_name, attr_name, default in self.ResponseDataParser:
            value = response.get(response_name)
            setattr(self, attr_name, value or default)

    def __repr__(self):
        return 'Vault: %s' % self.id

    @classmethod
    def create_vault(cls, api, name):
        api = APIProxy(api)
        response = api.create_vault(name)
        return Vault(api, api.describe_vault(response['x-oas-vault-id']))

    @classmethod
    def get_vault_by_id(cls, api, vault_id):
        api = APIProxy(api)
        return Vault(api, api.describe_vault(vault_id))

    @classmethod
    def get_vault_by_name(cls, api, vault_name):
        vaults = cls.list_all_vaults(api)
        for vault in vaults:
            if vault_name == vault.name:
                return vault
        raise ValueError('Vault not exists: %s' % vault_name)

    @classmethod
    def delete_vault_by_id(cls, api, vault_id):
        api = APIProxy(api)
        return api.delete_vault(vault_id)

    @classmethod
    def delete_vault_by_name(cls, api, vault_name):
        vaults = cls.list_all_vaults(api)
        for vault in vaults:
            if vault_name == vault.name:
                return vault.delete()
        raise ValueError('Vault not exists: %s' % vault_name)

    @classmethod
    def list_all_vaults(cls, api):
        api = APIProxy(api)
        result = api.list_vaults()
        return [Vault(api, data) for data in result['VaultList']]

    def list_all_multipart_uploads(self):
        result = self.api.list_multipart_uploads(self.id)
        return [Uploader(self, data) for data in result['UploadsList']]

    def list_all_jobs(self):
        result = self.api.list_jobs(self.id)
        return [Job(self, data) for data in result['JobList']]

    def delete(self):
        return self.api.delete_vault(self.id)

    def upload_archive(self, file_path, desc=None):
        length = os.path.getsize(file_path)
        if length > self.NormalUploadThreshold:
            return self.initiate_uploader(file_path, desc=desc).start()
        elif length > 0:
            return self._upload_archive_normal(file_path, desc=desc)
        else:
            raise ValueError('OAS does not support zero byte archive.')

    def _upload_archive_normal(self, file_path, desc):
        f = open_file(file_path=file_path)
        with f:
            file_size = content_length(f)
            etag = compute_etag_from_file(file_path)
            tree_etag = compute_tree_etag_from_file(file_path)
            mmapped_file = mmap.mmap(f.fileno(), length=file_size, offset=0,
                                     access=mmap.ACCESS_READ)
            try:
                response = self.api.upload_archive(self.id, mmapped_file,
                                                   etag, tree_etag,
                                                   size=file_size, desc=desc)
            finally:
                mmapped_file.close()
        return response['x-oas-archive-id']

    def initiate_uploader(self, file_path, desc=None):
        f = open_file(file_path=file_path)
        with f:
            size_total = content_length(f)
        part_size = Uploader.calc_part_size(size_total)

        response = self.api.initiate_multipart_upload(
            self.id, part_size, desc=desc)
        upload_id = response['x-oas-multipart-upload-id']

        response = self.api.describe_multipart(self.id, upload_id)
        return Uploader(self, response, file_path=file_path)

    def recover_uploader(self, upload_id):
        response = self.api.describe_multipart(self.id, upload_id)
        return Uploader(self, response)

    def delete_archive(self, archive_id):
        return self.api.delete_archive(self.id, archive_id)

    def get_job(self, job_id):
        response = self.api.describe_job(self.id, job_id)
        return Job(self, response)
    
    def pull_from_oss(self, osshost, bucket, object, desc=None):
        response = self.api.create_oss_transfer_job(
                  self.id, 'pull-from-oss', osshost, bucket, object, desc=desc)    
        return self.get_job(response['x-oas-job-id'])

    def push_to_oss(self, archive_id, osshost, bucket, object, desc=None):
        response = self.api.create_oss_transfer_job(
                  self.id, 'push-to-oss', osshost, bucket, object, archive_id=archive_id, desc=desc)    
        return self.get_job(response['x-oas-job-id'])

    def retrieve_archive(self, archive_id, desc=None, byte_range=None):
        response = self.api.create_job(
            self.id, 'archive-retrieval', archive_id=archive_id, desc=desc,
            byte_range='%d-%d' % byte_range
            if byte_range is not None else None)
        return self.get_job(response['x-oas-job-id'])

    def retrieve_inventory(self, desc=None):
        response = self.api.create_job(
            self.id, 'inventory-retrieval', desc=desc)
        return self.get_job(response['x-oas-job-id'])



import hashlib
import mmap
import os


class MerkleTree(object):

    def __init__(self, hash_list=None, hash_func=hashlib.md5):
        self.hash_list = hash_list or []
        self.hash_func = hash_func

    def append(self, block_hash):
        self.hash_list.append(block_hash)
        return self

    def load(self, hash_list):
        self.hash_list = hash_list
        return self

    def dump(self, delimiter=' '):
        return delimiter.join(self.hash_list)

    def digest(self):
        if len(self.hash_list) < 1:
            return ''

        hashes = list(self.hash_list)
        while len(hashes) > 1:
            new_hashes = list()

            while True:
                if len(hashes) > 1:
                    md = self.hash_func()
                    md.update(hashes.pop(0))
                    md.update(hashes.pop(0))
                    new_hashes.append(md.hexdigest().upper())
                elif len(hashes) > 0:
                    new_hashes.append(hashes.pop(0))
                else:
                    break

            hashes = new_hashes

        return hashes[0]


class TreeHashGenerator(object):

    MEGA_BYTE = 1024 * 1024

    def __init__(self, block_size=MEGA_BYTE, hash_func=hashlib.md5):
        self.block_size = block_size
        self.hash_func = hash_func
        self.stream = hash_func()
        self.remain = block_size
        self.tree = MerkleTree(hash_func=hash_func)

    def update(self, data, offset=0, length=None):
        length = length or len(data)
        while length > 0:
            if length >= self.remain:
                self.stream.update(data[offset:offset + self.remain])
                self.tree.append(self.stream.hexdigest().upper())
                self.stream = self.hash_func()
                offset += self.remain
                length -= self.remain
                self.remain = self.block_size
            else:
                self.stream.update(data[offset:offset + length])
                self.remain -= length
                length = 0

    def generate(self):
        if self.remain != self.block_size:
            self.tree.append(self.stream.hexdigest().upper())
            self.stream = self.hash_func()

        result = self.tree

        self.remain = self.block_size
        self.tree = MerkleTree(hash_func=self.hash_func)

        return result



# -*- coding: utf-8 -*-
# Copyright (C) 2011  Alibaba Cloud Computing
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.
#
import json

from httplib import HTTPException


class OASServerError(Exception):

    def __init__(self, response):
        raw_headers = response.getheaders()
        headers = dict()
        for k, v in raw_headers:
            headers[k.lower()] = v

        self.request_id = headers.get('x-oas-request-id')
        self.status = response.status

        content = ''
        try:
            content = response.read()
            body = json.loads(content)
            self.code = body.get('code')
            self.type = body.get('type')
            self.message = body.get('message')
            msg = 'Expected 2xx, got '
            msg += '(%d, code=%s, message=%s, type=%s, request_id=%s)' % \
                   (self.status, self.code,
                    self.message, self.type, self.request_id)
        except (HTTPException, ValueError):
            msg = 'Expected 2xx, got (%d, content=%s, request_id=%s)' % \
                  (self.status, content, self.request_id)

        super(OASServerError, self).__init__(msg)


class OASClientError(Exception):
    pass


class UploadArchiveError(OASClientError):
    pass


class DownloadArchiveError(OASClientError):
    pass


class HashDoesNotMatchError(OASClientError):
    pass



# -*- coding=UTF-8 -*-
# Copyright (C) 2011  Alibaba Cloud Computing
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.
#
import json
import httplib
import socket
import string
import sys
import time

from oas.oas_util import get_assign, append_param


class OASAPI(object):
    DefaultContentType = 'application/octet-stream'
    DefaultSendBufferSize = 8192
    DefaultGetBufferSize = 1024 * 1024 * 10
    provider = 'OAS'

    def __init__(self, host, accessId, accessKey, port=80, is_security=False):
        self.host = host
        self.port = port
        self.is_security = is_security
        self.access_id = accessId
        self.secret_access_key = accessKey

    def __get_connection(self):
        if self.is_security or self.port == 443:
            self.is_security = True
            return httplib.HTTPSConnection(host=self.host, port=self.port, timeout=100)
        else:
            return httplib.HTTPConnection(host=self.host, port=self.port, timeout=100)

    def __get_resource(self, params=None):
        if not params:
            return ''
        tmp_headers = {}
        for k, v in params.items():
            tmp_k = k.lower().strip()
            tmp_headers[tmp_k] = v
        override_response_list = [
            'limit', 'marker', 'response-content-type', 'response-content-language',
            'response-cache-control', 'logging', 'response-content-encoding',
            'acl', 'uploadId', 'uploads', 'partNumber', 'group',
            'delete', 'website', 'location', 'objectInfo',
            'response-expires', 'response-content-disposition']
        override_response_list.sort()
        resource = ''
        separator = '?'
        for i in override_response_list:
            if tmp_headers.has_key(i.lower()):
                resource += separator
                resource += i
                tmp_key = str(tmp_headers[i.lower()])
                if len(tmp_key) != 0:
                    resource += '='
                    resource += tmp_key
                separator = '&'
        return resource

    def __http_request(self, method, url, headers=None, body='', params=None):
        headers = headers or dict()
        headers['Date'] = time.strftime(
            '%a, %d %b %Y %H:%M:%S GMT', time.gmtime())
        headers['User-Agent'] = 'OAS Python SDK'
        headers['Host'] = self.host
        headers['x-oas-version'] = '0.2.3'
        if len(body) > 0:
            headers['Content-Length'] = len(body)
        resource = url + self.__get_resource(params)
        if params is not None:
            url = append_param(url, params)
        headers['Authorization'] = self.__create_sign_for_normal_auth(
            method, headers, resource)
        conn = self.__get_connection()
        try:
            conn.request(method, url, body, headers)
            return conn.getresponse()
        except socket.timeout, e:
            raise Exception('Connect or send timeout! ' + e.message)

    def __http_reader_huge_cache_request(self, method, url, headers, content):
        try:
            conn = self.__get_connection()
            resource = url
            conn.putrequest(method, url)
            for k in headers.keys():
                conn.putheader(str(k), str(headers[k]))
            if '' != self.secret_access_key and '' != self.access_id:
                auth = self.__create_sign_for_normal_auth(
                    method, headers, resource)
                conn.putheader('Authorization', auth)

            conn.endheaders()
            send_buffer_size = 1024 * 1024
            sended_size = 0
            content_length = len(content)
            left_length = content_length
            while True:
                if sended_size == content_length:
                    break
                elif sended_size > content_length:
                    raise Exception(
                        'Sended data more than content_length set.')

                left_length = content_length - sended_size
                if left_length > send_buffer_size:
                    #buf = reader.read(send_buffer_size)
                    buf = content[sended_size:(send_buffer_size+sended_size)]
                else:
                    #buf = reader.read(left_length)
                    buf = content[sended_size:content_length]
                buf_len = len(buf)
                if buf_len > 0:
                    conn.send(buf)
                    sended_size = sended_size + buf_len
                else:
                    break

            if sended_size < content_length:
                raise Exception('Sended data less than content_length set.')

            return conn.getresponse()

        except socket.timeout, e:
            raise Exception('Connect or send timeout! ' + e.message)
        except socket.error, e:
            error_info = str(e)
            bpipe_error = '[Errno 32] Broken pipe'
            if string.find(error_info, bpipe_error) >= 0:
                return conn.getresponse()
            else:
                raise Exception('Request error! ' + str(e))

    def __http_reader_request(self, method, url, headers, reader, content_length):
        try:
            conn = self.__get_connection()
            resource = url
            conn.putrequest(method, url)
            for k in headers.keys():
                conn.putheader(str(k), str(headers[k]))
            if '' != self.secret_access_key and '' != self.access_id:
                auth = self.__create_sign_for_normal_auth(
                    method, headers, resource)
                conn.putheader('Authorization', auth)

            conn.endheaders()
            send_buffer_size = 1024 * 1024
            sended_size = 0
            left_length = content_length
            while True:
                if sended_size == content_length:
                    break
                elif sended_size > content_length:
                    raise Exception(
                        'Sended data more than content_length set.')

                left_length = content_length - sended_size
                if left_length > send_buffer_size:
                    buf = reader.read(send_buffer_size)
                else:
                    buf = reader.read(left_length)
                buf_len = len(buf)
                if buf_len > 0:
                    conn.send(buf)
                    sended_size = sended_size + buf_len
                else:
                    break

            if sended_size < content_length:
                raise Exception('Sended data less than content_length set.')

            return conn.getresponse()

        except socket.timeout, e:
            raise Exception('Connect or send timeout! ' + e.message)
        except socket.error, e:
            error_info = str(e)
            bpipe_error = '[Errno 32] Broken pipe'
            if string.find(error_info, bpipe_error) >= 0:
                return conn.getresponse()
            else:
                raise Exception('Request error! ' + str(e))

    def __create_sign_for_normal_auth(self, method, headers=None, resource='/'):
        '''
        NOT public API
        Create the authorization for BC based on header input.
        it should be put into "Authorization" parameter of header.

        :type method: string
        :param:one of PUT, GET, DELETE, HEAD

        :type headers: dict
        :param: HTTP header

        :type resource: string
        :param:path of bucket or object, eg: /bucket/ or /bucket/object

        Returns:
            signature string
        '''
        auth_value = '%s %s:%s' % (self.provider, self.access_id, get_assign(
            self.secret_access_key, method, headers, resource))
        return auth_value

    def create_vault(self, vault_name):
        url = '/vaults/%s' % (vault_name)
        method = 'PUT'
        return self.__http_request(method, url)

    def delete_vault(self, vault_id):
        url = '/vaults/%s' % (vault_id)
        method = 'DELETE'
        return self.__http_request(method, url)

    def list_vault(self, marker=None, limit=None):
        '''
            @param marker: not required , string, marker of start
            @param limit:  not required , int, retrive size
        '''
        url = '/vaults'
        method = 'GET'
        params = dict()
        if marker is not None:
            params['marker'] = marker
        if limit is not None:
            params['limit'] = limit
        return self.__http_request(method, url, params=params)

    def get_vault_desc(self, vault_id):
        url = '/vaults/%s' % (vault_id)
        method = 'GET'
        return self.__http_request(method, url)

    def post_archive(self, vault_id, content, etag, tree_etag, desc=None):
        '''
            Post content to oas as archive
            @param content: required , string , content of archive
        '''
        url = '/vaults/%s/archives' % vault_id
        method = 'POST'
        headers = dict()
        headers['Host'] = self.host
        headers['x-oas-content-etag'] = etag
        headers['x-oas-tree-etag'] = tree_etag
        if desc is not None:
            headers['x-oas-archive-description'] = desc
        date = time.strftime('%a, %d %b %Y %H:%M:%S GMT', time.gmtime())
        headers['Content-Length'] = len(content)
        headers['Date'] = date

        if(len(content) < 512*1024*1024):
            return self.__http_request(method, url, headers, content)
        else:
            return self.__http_reader_huge_cache_request(method, url, headers, content)

    def post_archive_from_reader(self, vault_id, reader, content_length, etag, tree_etag, desc=None):
        '''
            Post archive from reader to oas
            @param content_length: required , int , byte length of archive
        '''
        url = '/vaults/%s/archives' % vault_id
        method = 'POST'
        headers = dict()
        headers['Host'] = self.host
        headers['x-oas-content-etag'] = etag
        headers['x-oas-tree-etag'] = tree_etag
        if desc is not None:
            headers['x-oas-archive-description'] = desc
        date = time.strftime('%a, %d %b %Y %H:%M:%S GMT', time.gmtime())
        headers['Content-Length'] = content_length
        headers['Date'] = date

        return self.__http_reader_request(method, url, headers, reader, content_length)

    def delete_archive(self, vault_id, archive_id):
        url = '/vaults/%s/archives/%s' % (vault_id, archive_id)
        method = 'DELETE'
        headers = {}
        res = self.__http_request(method, url, headers)
        return res

    def head_archive(self, vault_id, archive_id):
        url = '/vaults/%s/archives/%s/meta' % (vault_id, archive_id)
        method = 'GET'
        headers = {}
        res = self.__http_request(method, url, headers)
        return res

    def create_multipart_upload(self, vault_id, partsize, desc=None):
        '''
            @param partsize:required, int, size of per part should be large than 64M(67108864) and mod 1M equals 0
        '''
        url = '/vaults/%s/multipart-uploads' % (vault_id)
        headers = {}
        headers['x-oas-part-size'] = str(partsize)
        if desc != None:
            headers['x-oas-archive-description'] = desc
        method = 'POST'
        res = self.__http_request(method, url, headers)
        return res

    def list_multipart_upload(self, vault_id, marker=None, limit=None):
        '''
            list all multipart uploads
        '''
        url = '/vaults/%s/multipart-uploads' % (vault_id)
        method = 'GET'
        params = {}
        if marker is not None:
            params['marker'] = marker
        if limit is not None:
            params['limit'] = limit
        headers = {}
        body = ''
        res = self.__http_request(method, url, headers, body, params)
        return res

    def complete_multipart_upload(self, vault_id, upload_id, filesize, tree_etag):
        '''
            complete multipart upload
        '''
        url = '/vaults/%s/multipart-uploads/%s' % (vault_id, upload_id)
        method = 'POST'
        headers = {}
        headers['x-oas-archive-size'] = str(filesize)
        headers['x-oas-tree-etag'] = tree_etag
        res = self.__http_request(method, url, headers)
        return res

    def delete_multipart_upload(self, vault_id, upload_id):
        '''
            delete multipart upload
        '''
        url = '/vaults/%s/multipart-uploads/%s' % (vault_id, upload_id)
        method = 'DELETE'
        headers = {}
        res = self.__http_request(method, url, headers)
        return res

    def post_multipart(self, vault_id, upload_id, content, prange, etag, tree_etag):
        '''
            post content to oas as multipart

            @param content:required, string, upload data content
            @param prange: required, string, upload data range eg: 0-67108863
        '''
        url = '/vaults/%s/multipart-uploads/%s' % (vault_id, upload_id)
        method = 'PUT'
        headers = {}
        headers['Host'] = self.host
        headers['Content-Length'] = len(content)
        headers['x-oas-content-etag'] = etag
        headers['x-oas-tree-etag'] = tree_etag
        date = time.strftime('%a, %d %b %Y %H:%M:%S GMT', time.gmtime())
        headers['Date'] = date
        headers['Content-Range'] = prange

        if(len(content) < 512*1024*1024):
            return self.__http_request(method, url, headers, content)
        else:
            return self.__http_reader_huge_cache_request(method, url, headers, content)

    def post_multipart_from_reader(self, vault_id, upload_id, reader, content_length, prange, etag, tree_etag):
        '''
            post multipart from reader to oas

            @param content_length: size of post part , should be equal with the multipart upload setting,the last part can be less
            @param prange: required, string, upload data range eg: 0-67108863
        '''
        url = '/vaults/%s/multipart-uploads/%s' % (vault_id, upload_id)
        method = 'PUT'
        headers = {}
        headers['Host'] = self.host
        headers['Content-Length'] = content_length
        headers['x-oas-content-etag'] = etag
        headers['x-oas-tree-etag'] = tree_etag
        date = time.strftime('%a, %d %b %Y %H:%M:%S GMT', time.gmtime())
        headers['Date'] = date
        headers['Content-Range'] = prange

        return self.__http_reader_request(method, url, headers, reader, content_length)

    def list_multipart(self, vault_id, upload_id, marker=None, limit=None):
        '''
            list all multiparts belong to one upload
        '''
        url = '/vaults/%s/multipart-uploads/%s' % (vault_id, upload_id)
        method = 'GET'
        params = dict()
        if marker is not None:
            params['marker'] = marker
        if limit is not None:
            params['limit'] = limit
        return self.__http_request(method, url, params=params)

    def create_job(self, vault_id, job_type, archive_id=None, desc=None, byte_range=None):
        '''
            create job
            @param job_type: required, string, can only be archive-retrieval or inventory-retrieval
            @param archive_id: not required, string, when job_type is archive-retrieval , archive_id must be set
        '''
        url = '/vaults/%s/jobs' % (vault_id)
        method = 'POST'
        body = dict()
        body['Type'] = job_type
        if job_type == 'archive-retrieval':
            body['ArchiveId'] = archive_id
            if byte_range is not None:
                body['RetrievalByteRange'] = byte_range
        if desc is not None:
            body['Description'] = desc
        body = json.dumps(body)
        return self.__http_request(method, url, body=body)

    def create_oss_transfer_job(self, vault_id, job_type, osshost, bucket, object, archive_id=None, desc=None):
        '''
            create push-to-oss/pull-from-oss job
            @param job_type: required, string, can only be push-to-oss or pull-from-oss
            @param archiveid: not required, string, when job_type is push-to-oss, archive_id must be set
        '''
        url = '/vaults/%s/jobs' % (vault_id)
        method = 'POST'
        body = dict()
        body['Type'] = job_type
        body['OSSHost'] = osshost
        body['Bucket'] = bucket
        body['Object'] = object
        if archive_id is not None:
            body['ArchiveId'] = archive_id
        if desc is not None:
            body['Description'] = desc
        body = json.dumps(body)
        return self.__http_request(method, url, body=body)

    def get_jobdesc(self, vault_id, job_id):
        url = '/vaults/%s/jobs/%s' % (vault_id, job_id)
        method = 'GET'
        return self.__http_request(method, url)

    def fetch_job_output(self, vault_id, job_id, orange=None):
        '''
           fetch job output
           @param orange: not required , string, fetch byte range like bytes=0-1048575
        '''
        url = '/vaults/%s/jobs/%s/output' % (vault_id, job_id)
        method = 'GET'
        headers = dict()
        if orange is not None:
            headers['Range'] = orange
        return self.__http_request(method, url, headers)

    def list_job(self, vault_id, marker=None, limit=None):
        '''
            fetch all jobs
        '''
        url = '/vaults/%s/jobs' % (vault_id)
        method = 'GET'
        params = dict()
        if marker is not None:
            params['marker'] = marker
        if limit is not None:
            params['limit'] = limit
        return self.__http_request(method, url, params=params)



# -*- coding=UTF-8 -*-
# Copyright (C) 2011  Alibaba Cloud Computing
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.
#
import datetime
import json
import os
import sys
import time
import hashlib
from collections import namedtuple
import ConfigParser

from oas_api import OASAPI
from oas_util import UTC
from oas.ease.utils import *

OAS_PREFIX = 'oas://'

def convert_time_format(t):
    if not t:
        return ''
    t_datetime = datetime.datetime.strptime(t, '%a, %d %b %Y %H:%M:%S GMT')
    utc_timestamp = time.mktime(t_datetime.timetuple())
    utc8_datetime = datetime.datetime.fromtimestamp(utc_timestamp, UTC(16))
    return utc8_datetime.strftime('%Y-%m-%d %H:%M:%S')

def check_response(res):
    try:
        if res.status / 100 != 2:
            errmsg = ''
            errmsg += 'Error Headers:\n'
            errmsg += str(res.getheaders())
            errmsg += '\nError Body:\n'
            errmsg += res.read(1024)
            errmsg += '\nError Status:\n'
            errmsg += str(res.status)
            errmsg += '\nFailed!\n'
            sys.stderr.write(errmsg)
            sys.exit(1)
    except AttributeError, e:
        sys.stderr.write('Error: check response status failed! msg: %s\n' % e)
        sys.exit(1)

def parse_vault_name(path):
    if not path.lower().startswith(OAS_PREFIX):
        sys.stderr.write('oas vault path must start with %s\n' % OAS_PREFIX)
        sys.exit(1)
    path_fields = path[len(OAS_PREFIX):].split('/')
    name = path_fields[0]
    return name

class OASCMD(object):

    def __init__(self, auth_info):
        self.api = OASAPI(auth_info.host, auth_info.accessId,
                          auth_info.accessKey, port=auth_info.port)

    def byte_humanize(self, byte):
        if byte == None: return ''
        unitlist = ['B', 'K', 'M', 'G', 'T', 'P']
        unit = unitlist[0]
        for i in range(len(unitlist)):
            unit = unitlist[i]
            if int(byte) // 1024 == 0:
                break
            byte = float(byte) / 1024
        return '%.2f %s' % (byte, unit if unit == 'B' else unit+'B')

    def parse_size(self, size):
        try:
            if size == None: return size
            if isinstance(size, (int, long)): return size
            size = size.rstrip('B')
            if size.isdigit(): return int(size)
            byte, unit = int(size[:-1]), size[-1].upper()
            unitlist = ['B', 'K', 'M', 'G', 'T', 'P']
            if unit not in unitlist:
                byte = int(size)
            else:
                idx = unitlist.index(unit)
                for _ in range(idx): byte *= 1024
            return byte
        except Exception, e:
            sys.stderr.write('[Error]: Inapropriate size %s provided\n' % size)
            sys.stderr.write('Only numbers, optionally ended with K, M, G, ' \
                             'T, P are accepted, units are case insensitive\n')
            sys.exit(1)

    def kv_print(self, idict, title=None):
        keys = title or idict.keys()
        maxlen = max([len(t) for t in keys])
        fmt = u'{:>{lname}}: {}'
        for k in keys:
            print fmt.format(k, idict[k], lname=maxlen)

    def fetch_vault_id(self, vault_name):
        marker = ''
        while True:
            res = self.api.list_vault(marker, 10)
            check_response(res)
            response_content = res.read()
            rjson = json.loads(response_content, 'UTF8')
            marker = rjson['Marker']
            vault_list = rjson['VaultList']
            if vault_list:
                for vault_info in vault_list:
                    if vault_info['VaultName'] == vault_name:
                        return str(vault_info['VaultId'])

            if not marker:
                break
        sys.stderr.write("Error: vault with name '%s' not found\n" % vault_name)
        sys.exit(1)

    def cmd_ls(self, args):
        return self.cmd_listvault(args)

    def cmd_cv(self, args):
        return self.cmd_createvault(args)

    def cmd_rm(self, args):
        if not args.archiveId:
            return self.cmd_deletevault(args)
        return self.cmd_deletearchive(args)

    def cmd_upload(self, args):
        if not args.filename or not os.path.isfile(args.filename):
            sys.stderr.write("Error: file '%s' not existed\n" % args.filename)
            sys.exit(1)
        size = os.path.getsize(args.filename)
        desc = args.desc or args.filename[:128]
        if (not args.part_size and size >= 100*1024*1024) or \
                (args.part_size and size > 32*1024*1024):
            resume = 0
            if not args.uploadid:
                partsize = self.parse_size(args.part_size)
                if partsize:
                    if partsize % (32*1024*1024) != 0:
                        sys.stderr.write('Error: partsize must be divided by 32MB!\n')
                        sys.exit(1)
                    if partsize * 10000 < size:
                        print 'specified partsize too small, will be adjusted'
                    if partsize > size:
                        print 'specified partsize too large, will be adjusted'
                        while partsize > size:
                            partsize /= 2
                else:
                    print 'File larger than 100MB, multipart upload will be used'
                    partsize = 32*1024*1024
                while partsize * 10000 < size:
                    partsize *= 2
                nparts = size // partsize
                if size % partsize: nparts += 1
                print 'Use %s parts with partsize %s to upload' % \
                        (nparts, self.byte_humanize(partsize))

                Create = namedtuple('Namespace', ['vault', 'part_size', 'desc'])
                cargs = Create(args.vault, partsize, desc)
                uploadid = self.cmd_createmupload(cargs)
            else:
                uploadid = args.uploadid
                vault_name = parse_vault_name(args.vault)
                vaultId = self.fetch_vault_id(vault_name)
                res = self.api.list_multipart(vaultId, uploadid, None, None)
                check_response(res)
                rjson = json.loads(res.read(), 'UTF8')
                plist = rjson['Parts']
                partsize = rjson['PartSizeInBytes']
                nparts = size // partsize
                if size % partsize: nparts += 1
                if plist:
                    resumebytes = max([int(x['RangeInBytes'].split('-')[1]) \
                                       for x in plist]) + 1
                    if resumebytes == size:
                        resume = nparts
                    else:
                        resume = resumebytes // partsize
                print 'Resume last upload with partsize %s' % \
                        self.byte_humanize(partsize)
            Part = namedtuple('Namespace', ['vault', 'uploadid', 'filename',
                'start', 'end', 'etag', 'tree_etag'])
            start = 0
            etaglist = []
            for i in range(nparts):
                end = min(size, start+partsize)-1
                etag, tree_etag = compute_hash_from_file(args.filename, start,
                                                         end-start+1)
                pargs = Part(args.vault, uploadid, args.filename, start,
                             end, etag, tree_etag)
                start += partsize
                etaglist.append(tree_etag)
                if i < resume: continue
                print 'Uploading part %s...' % (i+1)
                self.cmd_postmpart(pargs)
            Complete = namedtuple('Namespace', ['vault', 'uploadid',
                                                'size', 'tree_etag'])
            etree = compute_combine_tree_etag_from_list(etaglist)
            cargs = Complete(args.vault, uploadid, size, etree)
            self.cmd_completemupload(cargs)
            return
        if size <= 32*1024*1204 and args.part_size:
            print 'File smaller than 32MB, part-size will be ignored.'
        Post = namedtuple('Namespace', ['vault', 'filename', 'etag',
                                        'tree_etag', 'desc'])
        etag, tree_etag = compute_hash_from_file(args.filename)
        pargs = Post(args.vault, args.filename, etag, tree_etag, desc)
        return self.cmd_postarchive(pargs)

    def cmd_fetch(self, args):
        return self.cmd_fetchjoboutput(args)

    def cmd_createvault(self, args):
        vault_name = parse_vault_name(args.vault)
        res = self.api.create_vault(vault_name)
        check_response(res)
        vault_id = res.getheader('x-oas-vault-id')
        print 'Vault ID: %s' % vault_id

    def cmd_deletevault(self, args):
        vault_name = parse_vault_name(args.vault)
        vault_id = self.fetch_vault_id(vault_name)
        res = self.api.delete_vault(vault_id)
        check_response(res)
        print 'Delete success'

    def cmd_listvault(self, args):
        marker = args.marker
        limit = args.limit
        res = self.api.list_vault(marker, limit)
        check_response(res)
        rjson = json.loads(res.read(), 'UTF8')
        marker = rjson['Marker']
        vault_list = rjson['VaultList']
        print 'Marker: %s' % marker
        print 'Vault count: %s' % len(vault_list)
        if vault_list:
            max_name = max(max([len(v['VaultName']) for v in vault_list]),
                           len('VaultName')) + 1
            fmt = u'{:<32} {:{lname}} {:<20} {:<16} {:<12} {:<20}'
            title = ('VaultID', 'VaultName', 'CreationDate', 'NumberOfArchives',
                     'TotalSize', 'LastInventoryDate')
            print fmt.format(*title, lname=max_name)
            print '-' * (105+max_name)
            for vault in vault_list:
                print fmt.format(
                     vault['VaultId'], vault['VaultName'],
                     convert_time_format(vault['CreationDate']),
                     vault['NumberOfArchives'],
                     self.byte_humanize(vault['SizeInBytes']),
                     convert_time_format(vault['LastInventoryDate']),
                     lname=max_name)

    def cmd_getvaultdesc(self, args):
        vault_name = parse_vault_name(args.vault)
        vault_id = self.fetch_vault_id(vault_name)
        res = self.api.get_vault_desc(vault_id)
        check_response(res)
        rjson = json.loads(res.read(), 'UTF8')
        title = ('VaultId', 'VaultName', 'CreationDate', 'NumberOfArchives',
                 'SizeInBytes', 'LastInventoryDate')
        for k in ('CreationDate', 'LastInventoryDate'):
            rjson[k] = convert_time_format(rjson[k])
        self.kv_print(rjson, title)

    def cmd_postarchive(self, args):
        vault_name = parse_vault_name(args.vault)
        vault_id = self.fetch_vault_id(vault_name)
        filepath = args.filename
        desc = args.desc or args.filename[:128]
        if not os.path.isfile(filepath):
            sys.stderr.write("Error: file '%s' not existed!\n" % filepath)
            sys.exit(1)

        file_size = os.path.getsize(filepath)
        etag = args.etag.upper() if args.etag else ''
        tree_etag = args.tree_etag.upper() if args.tree_etag else ''
        if not (etag and tree_etag):
            tmpsum, tmptree = compute_hash_from_file(args.filename)
            etag = etag or tmpsum
            tree_etag = tree_etag or tmptree
        with open(filepath, 'rb') as fp:
            res = self.api.post_archive_from_reader(
                vault_id, fp, file_size, etag, tree_etag, desc)
            check_response(res)
        archive_id = res.getheader('x-oas-archive-id')
        print 'Archive ID: %s' % archive_id

    def cmd_deletearchive(self, args):
        vault_name = parse_vault_name(args.vault)
        vault_id = self.fetch_vault_id(vault_name)
        res = self.api.delete_archive(vault_id, args.archiveId)
        check_response(res)
        print 'Delete success'

    def cmd_fileetag(self, args):
        if not os.path.isfile(args.filename):
            sys.stderr.write("Error: file '%s' not existed!\n" % args.filename)
            sys.exit(1)
        etag, tree_etag = compute_hash_from_file(args.filename)
        print "etag     :", etag
        print "tree_etag:", tree_etag

    def cmd_partetag(self, args):
        if not os.path.isfile(args.filename):
            sys.stderr.write("Error: file '%s' not existed!\n" % args.filename)
            sys.exit(1)
        start = self.parse_size(args.start)
        end = self.parse_size(args.end)
        if end % (1024*1024) == 0: end -= 1
        size = end - start + 1
        etag, tree_etag = compute_hash_from_file(args.filename, start, size)
        print "etag     :", etag
        print "tree_etag:", tree_etag

    def cmd_createmupload(self, args):
        vault_name = parse_vault_name(args.vault)
        vault_id = self.fetch_vault_id(vault_name)
        part_size = self.parse_size(args.part_size)
        desc = args.desc
        res = self.api.create_multipart_upload(vault_id, part_size, desc)
        check_response(res)
        upload_id = res.getheader('x-oas-multipart-upload-id')
        print 'MultiPartUpload ID: %s' % upload_id
        return upload_id

    def cmd_listmupload(self, args):
        vault_name = parse_vault_name(args.vault)
        vault_id = self.fetch_vault_id(vault_name)
        marker = args.marker
        limit = args.limit
        res = self.api.list_multipart_upload(vault_id, marker, limit)
        check_response(res)
        rjson = json.loads(res.read(), 'UTF8')
        marker = rjson['Marker']
        upload_list = rjson['UploadsList']
        print 'Marker: %s' % marker
        print 'Multipart upload count: %s' % len(upload_list)
        if upload_list:
            print
            maxdesc = max(max([len(m['ArchiveDescription']) for m in upload_list]),
                          len('ArchiveDescription')) + 1
            fmt = u'{:<32} {:<20} {:<10} {:<{ldesc}}'
            title = ('MultipartUploadId', 'CreationDate', 'PartSize',
                     'ArchiveDescription')
            print fmt.format(*title, ldesc=maxdesc)
            print '-' * (64+maxdesc)
            for upload in upload_list:
                print fmt.format(
                     upload['MultipartUploadId'],
                     convert_time_format(upload['CreationDate']),
                     self.byte_humanize(upload['PartSizeInBytes']),
                     upload['ArchiveDescription'],
                     ldesc=maxdesc)
            print

    def cmd_completemupload(self, args):
        vault_name = parse_vault_name(args.vault)
        vault_id = self.fetch_vault_id(vault_name)
        upload_id = args.uploadid
        file_size = self.parse_size(args.size)
        tree_etag = args.tree_etag.upper() if args.tree_etag else ''

        res = self.api.complete_multipart_upload(
            vault_id, upload_id, file_size,  tree_etag)
        check_response(res)
        archive_id = res.getheader('x-oas-archive-id')
        print 'Archive ID: %s' % archive_id

    def cmd_deletemupload(self, args):
        vault_name = parse_vault_name(args.vault)
        vault_id = self.fetch_vault_id(vault_name)
        res = self.api.delete_multipart_upload(vault_id, args.uploadid)
        check_response(res)
        print 'Delete success'

    def cmd_postmpart(self, args):
        vault_name = parse_vault_name(args.vault)
        vault_id = self.fetch_vault_id(vault_name)
        upload_id = args.uploadid
        filepath = args.filename
        start = self.parse_size(args.start)

        if not filepath or not os.path.isfile(filepath):
            sys.stderr.write("Error: file '%s' not existed!\n" % filepath)
            sys.exit(1)

        totalsize = os.path.getsize(filepath)
        end = args.end
        size = long(end) - long(start) + 1
        etag = args.etag.upper() if args.etag else None
        tree_etag = args.tree_etag.upper() if args.tree_etag else None
        if not (etag and tree_etag):
            tmpsum, tmptree = compute_hash_from_file(args.filename, start, size)
            etag = etag or tmpsum
            tree_etag = tree_etag or tmptree
        with open(filepath, 'rb') as file_reader:
            prange = '%s-%s' % (start, end)
            file_reader.seek(long(start))
            res = self.api.post_multipart_from_reader(
                vault_id, upload_id, file_reader, size,
                prange, etag, tree_etag)
            check_response(res)
        print 'Upload success'

    def cmd_listmpart(self, args):
        vault_name = parse_vault_name(args.vault)
        vault_id = self.fetch_vault_id(vault_name)
        upload_id = args.uploadid
        marker = args.marker
        limit = args.limit
        res = self.api.list_multipart(vault_id, upload_id, marker, limit)
        check_response(res)
        rjson = json.loads(res.read(), 'UTF8')
        rjson['CreationDate'] = convert_time_format(rjson['CreationDate'])
        title = ('MultipartUploadId', 'CreationDate', 'PartSizeInBytes',
                 'ArchiveDescription', 'Marker')
        self.kv_print(rjson, title)
        part_list = rjson['Parts']
        print '-' * 52
        print 'Part count: %s' % len(part_list)
        if len(part_list) > 0:
            print 'Part Ranges: '
            maxlen = max([len(p['RangeInBytes']) for p in part_list])
            fmt = u'{:>{lrange}}: {}'
            for part in part_list:
                print fmt.format(part['RangeInBytes'], part['ContentEtag'],
                                 lrange=maxlen)

    def cmd_createjob(self, args):
        vault_name = parse_vault_name(args.vault)
        vault_id = self.fetch_vault_id(vault_name)
        archive_id = args.archiveId
        desc = args.desc
        start = self.parse_size(args.start)
        size = self.parse_size(args.size)

        jtype = 'archive-retrieval' if args.archiveId else 'inventory-retrieval'
        if jtype == 'inventory-retrieval':
            if start or size:
                print 'Tip: Inventory-retrieval does NOT support range, ignored'
                start = size = None

        byte_range = None
        if (start != None and size != None):
            byte_range = '%s-%s' % (start, start+size-1)
        elif (start != None and size == None):
            res = self.api.head_archive(vault_id, archive_id)
            check_response(res)
            totalsize = json.loads(res.read(), 'UTF8')['ArchiveSizeInBytes']
            byte_range = '%s-%s' % (start, totalsize-1)
        elif (start == None and size != None):
            byte_range = '0-%s' % (size-1)
        if byte_range:
            print 'Archive retrieval range: %s' % byte_range

        res = self.api.create_job(
            vault_id, jtype, archive_id, desc, byte_range)
        check_response(res)
        job_id = res.getheader('x-oas-job-id')
        print '%s job created, job ID: %s' % (jtype, job_id)
        print 'Use\n\n    oascmd.py fetch %s %s <localfile>\n\nto check job ' \
              'progress and download the data when job finished' % \
              (args.vault, job_id)
        print 'NOTICE: Jobs usually take about 4 HOURS to complete.'
        return job_id

    def cmd_cp(self, args):
        src, dst = args.source, args.destination
        if not (src.startswith('oss://') and dst.startswith(OAS_PREFIX)) and \
                not (src.startswith(OAS_PREFIX) and dst.startswith('oss://')):
            sys.stderr.write('[Error]: Only paths startswith oss:// or ' \
                             '%s are supported\n' % OAS_PREFIX)
            return
        bucket = obj = vault_name = archiveId = None
        try:
            if src.startswith('oss://'):
                jtype  = 'pull-from-oss'
                ossstr = src[len('oss://'):]
                slash = ossstr.find('/')
                if slash == -1: raise ValueError
                desc   = args.desc or 'copy from %s' % ossstr
                desc   = desc[:128]
                bucket = ossstr[:slash]
                obj    = ossstr[slash+1:]
                vault_name = parse_vault_name(dst)
                vault_id = self.fetch_vault_id(vault_name)
            else:
                jtype  = 'push-to-oss'
                ossstr = dst[len('oss://'):]
                slash = ossstr.find('/')
                if slash == -1: raise ValueError
                desc   = args.desc or 'copy to %s' % ossstr
                desc   = desc[:128]
                bucket = ossstr[:slash]
                obj    = ossstr[slash+1:]
                vault_name, archiveId = src[len(OAS_PREFIX):].split('/')
                vault_id = self.fetch_vault_id(vault_name)
        except:
            sys.stderr.write('[Error]: Invalid source path or destination ' \
                    'path. \nOnly support OSS path like oss://bucketname/'  \
                    'objectname and OAS path like oas://vaultname or oas:'  \
                    '//vaultname/archiveId\n')
            return
        osshost = args.osshost
        if not osshost:
            try:
                config = ConfigParser.ConfigParser()
                config.read(os.path.expanduser('~/.osscredentials'))
                osshost = config.get('OSSCredentials', 'host')
                if not osshost: raise
            except:
                sys.stderr.write('OSS host not specified, nor can be read ' \
                    'from ~/.osscredentials\n')
                return

        res = self.api.create_oss_transfer_job(vault_id,
                jtype, osshost, bucket, obj, archive_id=archiveId, desc=desc)
        check_response(res)
        job_id = res.getheader('x-oas-job-id')
        print 'oss transfer job created, job ID: %s' % (job_id)
        tip = '.'
        if jtype == 'pull-from-oss':
            tip = ' and retrieve corresponding Archive ID from job status'
        print 'Use\n\n    oascmd.py getjobdesc oas://%s %s\n\nto check job ' \
              'progress%s' % (vault_name, job_id, tip)
        print 'NOTICE: Jobs usually take about 4 HOURS to complete.'
        return job_id

    def cmd_getjobdesc(self, args):
        vault_name = parse_vault_name(args.vault)
        vault_id = self.fetch_vault_id(vault_name)
        job_id = args.jobid
        res = self.api.get_jobdesc(vault_id, job_id)
        check_response(res)
        rjson = json.loads(res.read(), 'UTF8')

        title = ('JobId', 'Action', 'StatusCode', 'StatusMessage',
                'ArchiveId', 'ArchiveSizeInBytes',
                'TreeEtag', 'ArchiveTreeEtag',
                'RetrievalByteRange', 'Completed', 'CompletionDate',
                'CreationDate', 'InventorySizeInBytes', 'JobDescription')
        rjson['CreationDate'] = convert_time_format(rjson['CreationDate'])
        rjson['CompletionDate'] = convert_time_format(rjson['CompletionDate'])
        self.kv_print(rjson, title)

    def cmd_fetchjoboutput(self, args):
        vault_name = parse_vault_name(args.vault)
        vault_id = self.fetch_vault_id(vault_name)
        job_id = args.jobid
        dst = args.localfile
        start = self.parse_size(args.start)
        size = self.parse_size(args.size)

        vaultid = self.fetch_vault_id(parse_vault_name(args.vault))
        res = self.api.get_jobdesc(vaultid, args.jobid)
        check_response(res)
        job = json.loads(res.read(), 'UTF8')
        status = job['StatusCode'].lower()
        jtype = 'inventory-retrieval' if job['Action'] == 'InventoryRetrieval' \
                else 'archive-retrieval'
        if status == 'inprogress':
            print '%s job still in progress. Repeat this ' \
                  'command later' % jtype
            sys.exit(0)
        elif status == 'failed':
            print '%s job failed.' % jtype
            sys.exit(0)
        brange = None
        if jtype == 'archive-retrieval':
            brange = [int(x) for x in job['RetrievalByteRange'].split('-')]
        else:
            brange = [0, int(job['InventorySizeInBytes']) - 1]

        output_range = None
        if (start != None and size == None):
            size = brange[1] - brange[0] - start + 1
        elif (start == None and size != None):
            start = 0
        if (start != None and size != None):
            output_range = 'bytes=%s-%s' % (start, start+size-1)

        if not args.force and os.path.exists(dst):
            ans = raw_input('Output file %s existed. Do you wish to ' \
                    'overwrite it? (y/n): ' % dst)
            if ans.strip().lower() != 'y':
                print 'Answer is no. Quit now.'
                sys.exit(0)

        res = self.api.fetch_job_output(vault_id, job_id, output_range)
        check_response(res)

        with open(dst, 'wb') as f:
            total_read = 0
            while True:
                data = res.read(1024 * 1024)
                if len(data) > 0:
                    f.write(data)
                    total_read += len(data)
                    if jtype == 'inventory-retrieval':
                        continue
                    totalsize = size or brange[1] - brange[0] + 1
                    percent = total_read * 100 // totalsize
                    nbar = percent // 2
                    sys.stdout.write('\r')
                    msgbar = '[%s] %s%%' % ('='*nbar+'>'+' '*(50-nbar), percent)
                    sys.stdout.write(msgbar)
                    sys.stdout.flush()
                else:
                    break
        sys.stdout.write('\n')
        print 'Download job output success'

    def cmd_listjob(self, args):
        vault_name = parse_vault_name(args.vault)
        vault_id = self.fetch_vault_id(vault_name)
        marker = args.marker
        limit = args.limit
        res = self.api.list_job(vault_id, marker, limit)
        check_response(res)
        rjson = json.loads(res.read(), 'UTF8')
        marker = rjson['Marker']
        job_list = rjson['JobList']
        print 'Marker: %s' % marker
        print 'Job count: %s' % len(job_list)
        if not job_list:
            return
        print
        title = ('JobId', 'Action', 'StatusCode', 'StatusMessage',
                'ArchiveId', 'ArchiveSizeInBytes',
                'TreeEtag', 'ArchiveTreeEtag',
                'RetrievalByteRange', 'Completed', 'CompletionDate',
                'CreationDate', 'InventorySizeInBytes', 'JobDescription')
        for job in job_list:
            print '================ JobId: %s ================' % job['JobId']
            job['CreationDate'] = convert_time_format(job['CreationDate'])
            job['CompletionDate'] = convert_time_format(job['CompletionDate'])
            self.kv_print(job, title)
            print




#!/usr/bin/env python2.7
# -*- coding=UTF-8 -*-
# Copyright (C) 2011  Alibaba Cloud Computing
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.
#
import ConfigParser
import os
import sys
import time

from collections import namedtuple
from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter

from oas.oas_ops import OASCMD

AuthInfo = namedtuple(
    'AuthInfo', ['host', 'port', 'accessId', 'accessKey'], verbose=False)

DEFAULT_HOST = 'cn-hangzhou.oas.aliyuncs.com'
DEFAULT_PORT = 80
DEFAULT_CONFIGFILE = os.path.expanduser('~') + '/.oascredentials'
CONFIGSECTION = 'OASCredentials'
HELP = \
    '''Usage: oascmd <action> [<args>]:

General Use:
    ls
    cv                   oas://vault
    rm                   oas://vault [archive_id]
    upload               oas://vault filename [-p PART_SIZE] [--desc desc]
    createjob            oas://vault [archive_id] [--start start] [--size size] [--desc desc]
    fetch                oas://vault jobid localfile [--start start] [--size size] [-f]

Vault Operations:
    createvault          oas://vault
    deletevault          oas://vault
    listvault            [--marker marker] [--limit limit]
    getvaultdesc         oas://vault

Archive Operations:
    postarchive          oas://vault local_file [etag] [tree_etag] [--desc desc]
    deletearchive        oas://vault archive_id

Etag Operations:
    fileetag             local_file
    partetag             local_file start end

Multipart Archive Operations:
    createmupload        oas://vault part_size [--desc desc]
    listmupload          oas://vault [--marker marker] [--limit limit]
    completemupload      oas://vault upload_id file_size [tree_etag]
    deletemupload        oas://vault upoad_id
    postmpart            oas://vault upload_id local_file start end [etag] [tree_etag]
    listmpart            oas://vault upload_id [--maker marker] [--limit limit]

Job Operations:
    createjob            oas://vault [archive_id] [--desc desc] [--start start] [--size size]
    createclodbackjob    oas://vault [--desc desc] host bucket object
    createdefrozenjob    oas://vault [--desc desc] host bucket object
    getjobdesc           oas://vault job_id
    fetchjoboutput       oas://vault job_id local_file [--start start] [--size size]
    listjob              oas://vault [--marker marker] [--limit limit]

Other Operations:
    config               [--host host] [-p,--port port] -i,--id=accessid -k,--key=accesskey
    help

'''

def fetch_credentials(args):
    try:
        host, port = args.host, args.port
        accessid, accesskey = args.id, args.key
        if not (host and port and accessid and accesskey):
            config = ConfigParser.ConfigParser()
            cfgfile = args.config_file or DEFAULT_CONFIGFILE
            config.read(cfgfile)
            # user defined inputs have higher priority
            host = host or config.get(CONFIGSECTION, 'host')
            port = port or int(config.get(CONFIGSECTION, 'port'))
            accessid = accessid or config.get(CONFIGSECTION, 'accessid')
            accesskey = accesskey or config.get(CONFIGSECTION, 'accesskey')
        return AuthInfo(host, port, accessid, accesskey)
    except:
        sys.stderr.write("Cannot get accessid/accesskey. " \
                         "Setup use: oascmd.py config\n")
        sys.exit(1)

def cmd_configure(args):
    config = ConfigParser.RawConfigParser()
    config.add_section(CONFIGSECTION)
    config.set(CONFIGSECTION, 'host', args.host)
    config.set(CONFIGSECTION, 'port', args.port)
    config.set(CONFIGSECTION, 'accessid', args.id)
    config.set(CONFIGSECTION, 'accesskey', args.key)
    # set config_file
    cfgfile = args.config_file or DEFAULT_CONFIGFILE
    if os.path.isfile(cfgfile):
        ans = raw_input('File existed. Do you wish to overwrite it?(y/n)')
        if ans.lower() != 'y':
            print 'Answer is No. Quit now'
            return
    with open(cfgfile, 'w+') as f:
        config.write(f)
    print 'Your configuration is saved to %s.' % cfgfile

def cmd_help(args):
    print HELP

def add_config(parser):
    parser.add_argument('--host', type=str, help='service host')
    parser.add_argument('--port', type=int, help='service port',
                        default=DEFAULT_PORT)
    parser.add_argument('-i', '--id', type=str, help='access ID')
    parser.add_argument('-k', '--key', type=str, help='access key')
    parser.add_argument('--config-file', type=str, help='configuration file')

if __name__ == '__main__':
    parser = ArgumentParser(formatter_class=ArgumentDefaultsHelpFormatter)
    subcmd = parser.add_subparsers(dest='cmd', title='Supported actions', \
            metavar='cmd', description=\
            """
            Commands {ls, cv, rm, upload, createjob, cp, fetch} provide easier
            ways to use OAS by combining commands below them. Generally they will
            suffice your daily use. For advanced operations, use commands like
            {createvault...}
            """)

    cmd = 'config'
    pcfg = subcmd.add_parser(cmd, help='config oascmd')
    pcfg.add_argument('--host', type=str, help='service host', required=True)
    pcfg.add_argument('-p', '--port', type=int, help='service port',
                      default=DEFAULT_PORT)
    pcfg.add_argument('-i', '--id', type=str, help='access ID', required=True)
    pcfg.add_argument('-k', '--key', type=str, help='access key', required=True)
    pcfg.add_argument('--config-file', type=str, help=\
            'file to save configuration')

    cmd = 'help'
    phelp = subcmd.add_parser(cmd, help='show a detailed help message and exit')

    cmd = 'ls'
    pls = subcmd.add_parser(cmd, help='list all vaults')
    pls.add_argument('--marker', type=str, help='to retrieve following list')
    pls.add_argument('--limit', type=int, help='number of vaults to be listed')
    add_config(pls)

    # only Python 3 supports aliases
    cmd = 'cv'
    pcv = subcmd.add_parser(cmd, help='create a vault')
    pcv.add_argument('vault', type=str, help='format oas://vault-name')
    add_config(pcv)

    cmd = 'rm'
    prm = subcmd.add_parser(cmd, help='remove a vault or an archive')
    prm.add_argument('vault', type=str, help='format oas://vault-name')
    prm.add_argument('archiveId', nargs='?', type=str, help=
            'ID of archive to be deleted')
    add_config(prm)

    cmd = 'upload'
    pupload = subcmd.add_parser(cmd, help='upload a local file')
    pupload.add_argument('vault', type=str, help='vault to store the file')
    pupload.add_argument('filename', type=str, help='file to be uploaded')
    pupload.add_argument('uploadid', nargs='?', type=str, help=\
            'MultiPartUpload ID upload returned to resume last upload')
    pupload.add_argument('--desc', type=str, help='description of the file')
    pupload.add_argument('-p', '--part-size', type=str, help=
            'multipart upload part size')
    add_config(pupload)

    cmd = 'createjob'
    pcj = subcmd.add_parser(cmd, help=\
            'create an inventory/archive retrieval job')
    pcj.add_argument('vault', type=str, help='vault to create the job')
    pcj.add_argument('archiveId', nargs='?', help='ID of archive to be ' \
            'downloaded. If not provided, an inventory-retrieval job will be ' \
            'created')
    pcj.add_argument('--start', help=\
            'start position of archive to retrieve, default to be 0')
    pcj.add_argument('--size', help=\
            'size to retrieve, default to be (totalsize - start)')
    pcj.add_argument('--desc', type=str, help='description of the job')
    add_config(pcj)

    cmd = 'cp'
    pcp = subcmd.add_parser(cmd, help='copy between OSS and OAS')
    pcp.add_argument('--osshost', type=str, help=\
            'OSS host address. Read from ~/.osscredentials if not provided')
    pcp.add_argument('source', type=str, help=\
            'source OSS object full path or OAS archive path')
    pcp.add_argument('destination', type=str, help=\
            'destination OSS object full path or full archive path')
    pcp.add_argument('--desc', type=str, help='description of the copy job')
    add_config(pcp)

    cmd = 'fetch'
    pfj = subcmd.add_parser(cmd, help='fetch job output')
    pfj.add_argument('vault', type=str, help='format oas://vault-name')
    pfj.add_argument('jobid', type=str, help='jobId createjob returned')
    pfj.add_argument('localfile', type=str, help='local file output written to')
    pfj.add_argument('-f', '--force', action='store_true', help=\
            'force overwrite if file exists')
    pfj.add_argument('--start', type=str, help=\
            'start position to download output retrieved, default to be 0')
    pfj.add_argument('--size', type=str, help=\
            'size to download, default to be (totalsize - start)')
    add_config(pfj)

    cmd = 'createvault'
    pcvault = subcmd.add_parser(cmd, help='create a vault')
    pcvault.add_argument('vault', type=str, help='format oas://vault-name')
    add_config(pcvault)

    cmd = 'deletevault'
    pdvault = subcmd.add_parser(cmd, help='delete a vault')
    pdvault.add_argument('vault', type=str, help='format oas://vault-name')
    add_config(pdvault)

    cmd = 'listvault'
    plsv = subcmd.add_parser(cmd, help='list all vaults')
    plsv.add_argument('--marker', type=str, help='to retrieve following list')
    plsv.add_argument('--limit', type=int, help='number of vaults to be listed')
    add_config(plsv)

    cmd = 'getvaultdesc'
    pgv = subcmd.add_parser(cmd, help='get detailed vault description')
    pgv.add_argument('vault', type=str, help='format oas://vault name')
    add_config(pgv)

    cmd = 'postarchive'
    ppost = subcmd.add_parser(cmd, help='upload a local file')
    ppost.add_argument('vault', type=str, help='vault to store the file')
    ppost.add_argument('filename', type=str, help='file to be uploaded')
    ppost.add_argument('etag', nargs='?', type=str, help='etag fileetag returned')
    ppost.add_argument('tree_etag', nargs='?', type=str, help=\
            'tree_etag fileetag returned')
    ppost.add_argument('--desc', type=str, help='description of the file')
    add_config(ppost)

    cmd = 'deletearchive'
    pdar = subcmd.add_parser(cmd, help='delete an archive')
    pdar.add_argument('vault', type=str, help='format oas://vault-name')
    pdar.add_argument('archiveId', type=str, help='ID of archive to be deleted')
    add_config(pdar)

    cmd = 'fileetag'
    pfetag = subcmd.add_parser(cmd, help='calculate tree etag of a file')
    pfetag.add_argument('filename', type=str, help='file to be calculated')
    add_config(pfetag)

    cmd = 'partetag'
    ppetag = subcmd.add_parser(cmd, help=\
            'calculate tree etag of a multipart upload part')
    ppetag.add_argument('filename', type=str, help='file to be read from')
    ppetag.add_argument('start', type=str, help='start position to read')
    ppetag.add_argument('end', type=str, help='end position to read')
    add_config(ppetag)

    cmd = 'createmupload'
    pcm = subcmd.add_parser(cmd, help='initiate a multipart upload')
    pcm.add_argument('vault', type=str, help='format oas://vault-name')
    pcm.add_argument('part_size', type=str, help='size of each multipart upload')
    pcm.add_argument('--desc', type=str, help='description of the upload')
    add_config(pcm)

    cmd = 'deletemupload'
    pdm = subcmd.add_parser(cmd, help='cancel a multipart upload')
    pdm.add_argument('vault', type=str, help='format oas://vault-name')
    pdm.add_argument('uploadid', type=str, help=
            'ID of multipart upload to be cancelled')
    add_config(pdm)

    cmd = 'listmupload'
    plm = subcmd.add_parser(cmd, help='list all multipart uploads in a vault')
    plm.add_argument('vault', type=str, help='format oas://vault-name')
    plm.add_argument('--marker', type=str, help='to retrieve following list')
    plm.add_argument('--limit', type=int, help='number to be listed')
    add_config(plm)

    cmd = 'postmpart'
    ppm = subcmd.add_parser(cmd, help='upload one part')
    ppm.add_argument('vault', type=str, help='vault to store the part')
    ppm.add_argument('uploadid', type=str, help='ID createmupload returned')
    ppm.add_argument('filename', type=str, help='file to read from')
    ppm.add_argument('start', type=str, help=
            'read start position, start must be divided by partsize')
    ppm.add_argument('end', type=str, help=
            'read end position, end+1 must be the size of file or '
            'partsize larger than start')
    ppm.add_argument('etag', nargs='?', help='etag partetag returned')
    ppm.add_argument('tree_etag', nargs='?', help='tree_etag partetag returned')
    add_config(ppm)

    cmd = 'listmpart'
    plp = subcmd.add_parser(cmd, help='list all parts uploaded in one upload')
    plp.add_argument('vault', type=str, help='format oas://vault-name')
    plp.add_argument('uploadid', type=str, help='ID of multipart upload')
    plp.add_argument('--marker', type=str, help='to retrieve following list')
    plp.add_argument('--limit', type=int, help='number to be listed')
    add_config(plp)

    cmd = 'completemupload'
    pcm = subcmd.add_parser(cmd, help='complete the multipart upload')
    pcm.add_argument('vault', type=str, help='vault where the upload initiated')
    pcm.add_argument('uploadid', type=str, help='ID createmupload returned')
    pcm.add_argument('size', type=str, help='size of the file')
    pcm.add_argument('tree_etag', type=str, help='tree_etag fileetag returned')
    add_config(pcm)

    cmd = 'getjobdesc'
    pgj = subcmd.add_parser(cmd, help='get job status')
    pgj.add_argument('vault', type=str, help='format oas://vault-name')
    pgj.add_argument('jobid', type=str, help='jobId createjob returned')
    add_config(pgj)

    cmd = 'fetchjoboutput'
    pfjob = subcmd.add_parser(cmd, help='fetch job output')
    pfjob.add_argument('vault', type=str, help='format oas://vault-name')
    pfjob.add_argument('jobid', type=str, help='jobId createjob returned')
    pfjob.add_argument('localfile', type=str, help='local file output written to')
    pfjob.add_argument('-f', '--force', action='store_true', help=\
            'force overwrite if file exists')
    pfjob.add_argument('--start', type=str, help=\
            'start position to download output retrieved, default to be 0')
    pfjob.add_argument('--size', type=str, help=\
            'size to download, default to be (totalsize - start)')
    add_config(pfjob)

    cmd = 'listjob'
    plj = subcmd.add_parser(cmd, help='list all jobs except expired')
    plj.add_argument('vault', type=str, help='format oas://vault-name')
    plj.add_argument('--marker', type=str, help=\
            'listjob returned to retrieve following list')
    plj.add_argument('--limit', type=int, help='number to be listed')
    add_config(plj)

    args = parser.parse_args()

    if args.cmd == 'help':
        cmd_help(args)
        sys.exit(0)
    elif args.cmd == 'config':
        cmd_configure(args)
        sys.exit(0)

    # fetch auth_info
    auth_info = fetch_credentials(args)

    # error command
    method = OASCMD.__dict__.get('cmd_%s' % args.cmd)
    if method == None:
        sys.stderr.write('Unsupported command: %s\nUse help for more ' \
                         'information\n' % args.cmd)
        sys.exit(1)

    # execute command
    begin = time.time()
    oas_cmd = OASCMD(auth_info)
    method(oas_cmd, args)
    end = time.time()
    sys.stderr.write('%.3f(s) elapsed\n' % (end - begin))



# -*- coding=UTF-8 -*-
# Copyright (C) 2011  Alibaba Cloud Computing
# 
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.
# 
import base64
import hashlib
from hashlib import sha1 as sha
import hmac
import urllib
from datetime import tzinfo, timedelta

SELF_DEFINE_HEADER_PREFIX = "x-oas-"


class UTC(tzinfo):

    """UTC"""

    def __init__(self, offset=0):
        self._offset = offset

    def utcoffset(self, dt):
        return timedelta(hours=self._offset)

    def tzname(self, dt):
        return "UTC +%s" % self._offset

    def dst(self, dt):
        return timedelta(hours=self._offset)


def safe_get_element(name, container):
    for k, v in container.items():
        if k.strip().lower() == name.strip().lower():
            return v
    return ""


def format_header(headers=None):
    '''
        format the headers that self define
        convert the self define headers to lower.
    '''
    if not headers:
        headers = {}
    tmp_headers = {}
    for k in headers.keys():
        if isinstance(headers[k], unicode):
            headers[k] = headers[k].encode('utf-8')

        if k.lower().startswith(SELF_DEFINE_HEADER_PREFIX):
            k_lower = k.lower()
            tmp_headers[k_lower] = headers[k]
        else:
            tmp_headers[k] = headers[k]
    return tmp_headers


def get_assign(secret_access_key, method, headers=None, resource="/", result=None):
    '''
        Create the authorization for BC based on header input.
        You should put it into "Authorization" parameter of header.
    '''
    if not headers:
        headers = {}
    if not result:
        result = []

    canonicalized_bc_headers = ""
    date = safe_get_element('Date', headers)
    canonicalized_resource = resource
    tmp_headers = format_header(headers)
    if len(tmp_headers) > 0:
        x_header_list = tmp_headers.keys()
        x_header_list.sort()
        for k in x_header_list:
            if k.startswith(SELF_DEFINE_HEADER_PREFIX):
                canonicalized_bc_headers += k + ":" + tmp_headers[k] + "\n"
    string_to_sign = method + "\n" + date + "\n" + \
        canonicalized_bc_headers + canonicalized_resource
    result.append(string_to_sign)
    h = hmac.new(secret_access_key, string_to_sign, sha)
    return base64.encodestring(h.digest()).strip()


def append_param(url, params):
    '''
        convert the parameters to query string of URI.
    '''
    l = []
    for k, v in params.items():
        k = k.replace('_', '-')
        if k == 'maxkeys':
            k = 'max-keys'
        if isinstance(v, unicode):
            v = v.encode('utf-8')
        if v is not None and v != '':
            l.append('%s=%s' % (urllib.quote(k), urllib.quote(str(v))))
        elif k == 'acl':
            l.append('%s' % (urllib.quote(k)))
        elif v is None or v == '':
            l.append('%s' % (urllib.quote(k)))
    if len(l):
        url = url + '?' + '&'.join(l)
    return url



# -*- coding: utf-8 -*-
# Copyright (C) 2011  Alibaba Cloud Computing
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.
#
import logging
import random
import threading
import time
import yaml
try:
    from collections import OrderedDict
except ImportError:
    from ordereddict import OrderedDict
from multiprocessing.pool import ThreadPool

from exceptions import *
from merkle import MerkleTree, TreeHashGenerator
from utils import *


log = logging.getLogger(__name__)


class Job(object):
    _MEGABYTE = 1024 * 1024

    NumberThread = 16
    NumRetry = 3

    ResponseDataParser = (('Action', 'action', None),
                          ('ArchiveTreeEtag', 'etag', None),
                          ('ArchiveId', 'archive_id', None),
                          ('ArchiveSizeInBytes', 'archive_size', 0),
                          ('Completed', 'completed', False),
                          ('CompletionDate', 'completion_date', None),
                          ('CreationDate', 'creation_date', None),
                          ('InventorySizeInBytes', 'inventory_size', 0),
                          ('JobDescription', 'description', None),
                          ('JobId', 'id', None),
                          ('RetrievalByteRange', 'job_range', None),
                          ('StatusCode', 'status_code', None),
                          ('StatusMessage', 'status_message', None),
                          ('TreeEtag', 'tree_etag', None))

    def __init__(self, vault, response):
        self.vault = vault
        self.parts = OrderedDict()
        self._update(response)

    def __repr__(self):
        return 'Job: %s' % self.id

    def _parse_etag(self):
        tag = self.etag.split('-')
        return tag[0], int(tag[1])

    def _is_multipart(self):
        return '-' in self.etag

    def _parse_job_range(self):
        return tuple([int(v) for v in self.job_range.split('-')])

    def _is_complete_download(self):
        byte_range = self._parse_job_range()
        return byte_range == (0, self.archive_size - 1)

    def _is_tree_hash_align(self):
        return self.tree_etag is not None

    def calc_part_size(self, size_total):
        part_size = 32*1024*1024

        number_parts = calc_num_part(part_size, size_total)
        if number_parts > 10000:
            part_size_refer = math.ceil(
                size_total / 10000 / (1024*1024))
            part_size_refer = int(part_size_refer * 1024*1024)
            part_bit_len = part_size_refer.bit_length() - 1
            part_size = 1<< part_bit_len
            while True:
                if(part_size >= part_size_refer):
                    break
                part_bit_len += 1
                part_size = 1<< part_bit_len 
        return part_size

    def _update(self, response):
        for response_name, attr_name, default in self.ResponseDataParser:
            value = response.get(response_name)
            setattr(self, attr_name, value or default)

        self.parts = OrderedDict()
        if self.archive_size > 0 and self.action == "ArchiveRetrieval":
            tmp_byte_range = self._parse_job_range()
            size = range_size(tmp_byte_range)
            if size > 32*1024*1024:
                part_size = self.calc_part_size(size)
                for byte_range in calc_ranges(part_size, size):
                    self.parts[byte_range] = None
            else:
                self.parts[(0, size - 1)] = None

    def update_status(self):
        response = self.vault.api.describe_job(self.vault.id, self.id)
        self._update(response)

    def _check_status(self, block=False):
        self.update_status()
        if not block:
            if not self.completed:
                raise DownloadArchiveError('Job not ready')
            elif self.status_code.lower() == 'failed':
                raise DownloadArchiveError('Job process failed')
        elif block:
            while not self.completed:
                log.info('Job :'+ self.id +' status: ' + self.status_code)
                time.sleep(random.randint(6, 9))
                self.update_status()

    def download_by_range(self, byte_range, file_path=None, file_obj=None,
                          chunk_size=None, block=True):
        if self.action == "PullFromOSS" or self.action == "PushToOSS":
            raise DownloadArchiveError('Job not ready')

        self._check_status(block)

        chunk_size = chunk_size or self._MEGABYTE
        f = open_file(file_path=file_path, file_obj=file_obj, mode='wb+')
        offset = f.tell() if file_obj is not None else 0
        size = range_size(byte_range)

        try:
            for cnt in xrange(self.NumRetry):
                pos = 0
                try:
                    response = self.vault.api.get_job_output(
                        self.vault.id, self.id, byte_range=byte_range)
                    while True:
                        data = response.read(chunk_size)
                        if not data:
                            break
                        f.write(data)
                        pos += len(data)

                    if pos == size:
                        return
                except OASServerError as e:
                    log.error('Range %d-%d download failed. OASServerError Reason: %s' %
                              (byte_range[0], byte_range[1], e))
                    if e.type != 'client':
                        f.seek(offset)
                        continue
                    else:
                        raise e
                except IOError as e:
                    log.error('Range %d-%d download failed. IOError Reason: %s' %
                              (byte_range[0], byte_range[1], e))
                    f.seek(offset)
                    continue
                except Exception as e:
                    log.error('Range %d-%d download failed. Exception Reason: %s' %
                              (byte_range[0], byte_range[1], e))
                    f.seek(offset)
                    continue
        finally:
            f.flush()
            if f is not file_obj:
                f.close()

        raise DownloadArchiveError(
            'Incomplete download: %d / %d' % (pos, size))

    def download_to_file(self, file_path, chunk_size=None, block=True):
        if self.action == "PullFromOSS" or self.action == "PushToOSS":
            raise DownloadArchiveError('Job not ready')

        self._check_status(block)
        chunk_size = chunk_size or self._MEGABYTE
        if self.inventory_size > 0:
            return self.download_by_range(
                file_path=file_path,
                byte_range=(0, self.inventory_size - 1))

        #if self._is_multipart():
        #    _, part_size = self._parse_etag()
        #else:
        #    _, part_size = self.etag, self.archive_size

        file_dir, file_name = os.path.split(file_path)
        log_file = os.path.join(file_dir, file_name + '.oas')
        try:
            self._load(log_file)
        except IOError:
            pass
        log_lock = threading.RLock()
        file_lock = threading.RLock()

        def download_part(byte_range):
            for cnt in xrange(self.NumRetry):
                time.sleep(random.randint(256, 4096) / 1000.)
                try:
                    response = self.vault.api.get_job_output(
                        self.vault.id, self.id, byte_range=byte_range)

                    generator = TreeHashGenerator()
                    offset = byte_range[0]
                    while True:
                        data = response.read(chunk_size)
                        if not data:
                            break
                        generator.update(data)

                        with file_lock:
                            f.seek(offset)
                            f.write(data)
                            f.flush()
                            offset += len(data)

                    if offset != byte_range[1] + 1:
                        log.error('Range %d-%d(%d) incomplete download.' %
                                  (byte_range[0], byte_range[1], offset))
                        continue
                    tree_etag = generator.generate().digest()
                    if self._is_tree_hash_align() and \
                            tree_etag != response['x-oas-tree-etag']:
                        log.error('Range %d-%d invalid checksum %s, '
                                  'which %s expected.' %
                                  (byte_range[0], byte_range[1], tree_etag,
                                   response['x-oas-tree-etag']))
                        continue
                    else:
                        self.parts[byte_range] = tree_etag
                        with log_lock:
                            self._save(log_file)
                        log.info('Range %d-%d download success.' % byte_range)
                        return
                except (OASServerError, IOError) as e:
                    log.error('Range %d-%d download failed. Reason: %s' %
                              (byte_range[0], byte_range[1], e))
                    continue
                except Exception as e:
                    log.error('Range %d-%d download failed. The reason: %s' %
                              (byte_range[0], byte_range[1], e))
                    continue

            log.info('Range %d-%d download failed.' % byte_range)

        f = open_file(file_path=file_path, mode='wb+')
        with f:
            log.info('Start download.')
            pool = ThreadPool(
                            processes=min(self.NumberThread, len(self.parts)))
            pool.map(download_part,
                     [byte_range for byte_range, tag in self.parts.items()
                      if tag is None])

            size = self.size_completed
            size_total = range_size(self._parse_job_range())
            if size != size_total:
                raise DownloadArchiveError(
                    'Incomplete download: %d / %d' %
                    (size, size_total))
            if self._is_tree_hash_align() == True:
                tree_etag_list = [tree_etag for _,tree_etag in self.parts.items()]
                tree_etag_actual = compute_combine_tree_etag_from_list(tree_etag_list)
                if tree_etag_actual != self.tree_etag:
                    raise HashDoesNotMatchError(
                            'tree-etag not match: %s / %s (actual)' %
                            (tree_etag_actual, self.tree_etag))

            if os.path.exists(log_file):
                os.remove(log_file)

        log.info('Download finish.')

    def _save(self, file_path):
        f = open_file(file_path=file_path, mode='wb+')
        with f:
            f.write(yaml.dump(self.parts, default_flow_style=True))

    def _load(self, file_path):
        f = open_file(file_path=file_path)
        with f:
            self.parts = yaml.load(f)

    @property
    def size_completed(self):
        size_list = [range_size(byte_range)
                     for byte_range,tree_etag in self.parts.items()
                     if tree_etag is not None]
        return sum(size_list)



# -*- coding: utf-8 -*-
# Copyright (C) 2011  Alibaba Cloud Computing
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.
#
import hashlib
import io
import math
import mmap
import os

from merkle import TreeHashGenerator
from merkle import MerkleTree


def is_file_like(obj):
    return callable(getattr(obj, 'read', None))


def content_length(content):
    if hasattr(content, '__len__'):
        return len(content)

    if hasattr(content, 'len'):
        return content.len

    if hasattr(content, 'fileno'):
        try:
            fileno = content.fileno()
        except io.UnsupportedOperation:
            pass
        else:
            return os.fstat(fileno).st_size - content.tell()

    if hasattr(content, 'getvalue'):
        return len(content.getvalue())

    raise ValueError('Unsupported content type')


def open_file(file_path=None, file_obj=None, mode='rb'):
    if file_path is None and file_obj is None:
        raise ValueError(
            'Either file_path or file_name should be provided.')
    if file_obj is not None:
        return file_obj
    else:
        try:
            return open(file_path, mode)
        except Exception:
            raise IOError('Failed to open file: %s' % file_path)


def range_size(byte_range):
    return byte_range[1] - byte_range[0] + 1


def calc_num_part(part_size, size_total):
    return int(math.ceil(float(size_total) / part_size))


def calc_ranges(part_size, size_total):
    result = []
    number_parts = calc_num_part(part_size, size_total)
    for n in xrange(number_parts):
        start = n * part_size
        end = (n + 1) * part_size - 1 \
            if n != number_parts - 1 else size_total - 1
        result.append((start, end))
    return result


def compute_etag_from_file(file_path, offset=0, size=None,
                                  chunk_size=1024 * 1024):
    with open(file_path, 'rb') as f:
        return compute_etag_from_file_obj(
            f, offset=offset, size=size, chunk_size=chunk_size)


def compute_etag_from_file_obj(file_obj, offset=0, size=None,
                                      chunk_size=1024 * 1024):
    etag = hashlib.md5()
     
    size = size or os.fstat(file_obj.fileno()).st_size - offset

    if size != 0 and offset % mmap.ALLOCATIONGRANULARITY == 0:
        target = mmap.mmap(file_obj.fileno(), length=size,
                           offset=offset,
                           access=mmap.ACCESS_READ)
    else:
        target = file_obj
        target.seek(offset)

    while size > 0:
        data = target.read(chunk_size)
        etag.update(data[:min(len(data), size)])
        size -= len(data)

    if target is file_obj:
        file_obj.seek(offset)
    else:
        target.close()
    s= etag.hexdigest().upper()
    return s


def compute_combine_etag(etag_list):
    etag = hashlib.md5()
    etag.update(''.join(etag_list))
    return etag.hexdigest().upper()

def compute_combine_tree_etag_from_list(tree_etag_list):
    merkletree = MerkleTree(hash_list=tree_etag_list)
    return merkletree.digest()

def compute_tree_etag_from_file(file_path, offset=0, size=None,
                                chunk_size=1024 * 1024):
    with open(file_path, 'rb') as f:
        return compute_tree_etag_from_file_obj(
            f, offset=offset, size=size, chunk_size=chunk_size)


def compute_tree_etag_from_file_obj(file_obj, offset=0, size=None,
                                    chunk_size=1024 * 1024):
    generator = TreeHashGenerator()

    size = size or os.fstat(file_obj.fileno()).st_size - offset

    if size != 0 and offset % mmap.ALLOCATIONGRANULARITY == 0:
        target = mmap.mmap(file_obj.fileno(), length=size,
                           offset=offset,
                           access=mmap.ACCESS_READ)
    else:
        target = file_obj
        target.seek(offset)

    while size > 0:
        data = target.read(chunk_size)
        generator.update(data[:min(len(data), size)])
        size -= len(data)

    if target is file_obj:
        file_obj.seek(offset)
    else:
        target.close()
    return generator.generate().digest()


def compute_hash_from_file(file_path, offset=0, size=None, chunk_size=1024*1024):
    with open(file_path, 'rb') as f:
        return compute_hash_from_file_obj(
            f, offset=offset, size=size, chunk_size=chunk_size)
        
def compute_hash_from_file_obj(file_obj, offset=0, size=None,
                                    chunk_size=1024 * 1024):
    etag = hashlib.md5()
    generator = TreeHashGenerator()

    size = size or os.fstat(file_obj.fileno()).st_size - offset

    if size != 0 and offset % mmap.ALLOCATIONGRANULARITY == 0:
        target = mmap.mmap(file_obj.fileno(), length=size,
                           offset=offset,
                           access=mmap.ACCESS_READ)
    else:
        target = file_obj
        target.seek(offset)

    while size > 0:
        data = target.read(chunk_size)
        generator.update(data[:min(len(data), size)])
        etag.update(data[:min(len(data), size)])
        size -= len(data)

    if target is file_obj:
        file_obj.seek(offset)
    else:
        target.close()
    return ( etag.hexdigest().upper(), generator.generate().digest())



# -*- coding: utf-8 -*-
# Copyright (C) 2011  Alibaba Cloud Computing
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.
#
from oas.oas_api import OASAPI

from exceptions import OASServerError
from response import OASResponse
from utils import *


class APIProxy(object):

    def __init__(self, api):
        self.api = api

    def __getattr__(self, name):
        methods = [
            'create_vault', 'delete_vault', 'list_vault', 'get_vault_desc',
            'post_archive', 'post_archive_from_reader', 'delete_archive', 'head_archive',
            'create_multipart_upload', 'list_multipart_upload',
            'complete_multipart_upload', 'delete_multipart_upload',
            'post_multipart', 'post_multipart_from_reader', 'list_multipart',
            'create_job', 'create_oss_transfer_job', 'get_jobdesc', 'fetch_job_output', 'list_job']

        transform = {'describe_vault': 'get_vault_desc',
                     'describe_job': 'get_jobdesc',
                     'describe_multipart': 'list_multipart',
                     'initiate_multipart_upload': 'create_multipart_upload',
                     'cancel_multipart_upload': 'delete_multipart_upload'}

        if name in transform:
            name = transform[name]

        if name not in methods:
            return object.__getattribute__(self, name)

        def wrapped(*args, **kwargs):
            try:
                res = OASAPI.__getattribute__(self.api, name)(*args, **kwargs)
            except Exception as e:
                raise IOError(str(e))

            if res.status / 100 == 2:
                return OASResponse(res)
            else:
                raise OASServerError(res)
        return wrapped

    def list_vaults(self):
        result = self.list_vault()
        marker = result['Marker']
        while marker:
            res = self.list_vault(marker=marker)
            result['VaultList'].extend(res['VaultList'])
            marker = res['Marker']
        result['Marker'] = None
        return result

    def list_multipart_uploads(self, vault_id):
        result = self.list_multipart_upload(vault_id)
        marker = result['Marker']
        while marker:
            res = self.list_multipart_upload(vault_id, marker=marker)
            result['UploadsList'].extend(res['UploadsList'])
            marker = res['Marker']
        result['Marker'] = None
        return result

    def list_parts(self, vault_id, upload_id):
        result = self.list_multipart(vault_id, upload_id)
        marker = result['Marker']
        while marker:
            res = self.list_multipart(vault_id, upload_id, marker=marker)
            result['Parts'].extend(res['Parts'])
            marker = res['Marker']
        result['Marker'] = None
        return result

    def list_jobs(self, vault_id):
        result = self.list_job(vault_id)
        marker = result['Marker']
        while marker:
            res = self.list_job(vault_id, marker=marker)
            result['JobList'].extend(res['JobList'])
            marker = res['Marker']
        result['Marker'] = None
        return result

    def upload_archive(self, vault_id, content, etag, tree_etag,
                       size=None, desc=None):
        if is_file_like(content):
            return self.post_archive_from_reader(
                vault_id, content, min(size, content_length(content)),
                etag, tree_etag, desc=desc)
        else:
            return self.post_archive(vault_id, content,
                                     etag, tree_etag, desc=desc)

    def upload_part(self, vault_id, upload_id, content, byte_range,
                    etag, tree_etag):
        if is_file_like(content):
            size_total = content_length(content)
            if range_size(byte_range) > size_total:
                raise ValueError('Byte range exceeded: %d-%d' % byte_range)
            return self.post_multipart_from_reader(
                vault_id, upload_id, content, range_size(byte_range),
                '%d-%d' % byte_range, etag, tree_etag)
        else:
            return self.post_multipart(
                vault_id, upload_id, content,
                '%d-%d' % byte_range, etag, tree_etag)

    def get_job_output(self, vault_id, job_id, byte_range=None):
        return self.fetch_job_output(vault_id, job_id,
                                     'bytes=%d-%d' % byte_range)



# -*- coding: utf-8 -*-
# Copyright (C) 2011  Alibaba Cloud Computing
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.
#
from __future__ import division

import logging
import random
import time
try:
    from collections import OrderedDict
except ImportError:
    from ordereddict import OrderedDict
from multiprocessing.pool import ThreadPool

from exceptions import *
from merkle import *
from utils import *

log = logging.getLogger(__name__)


class Uploader(object):
    _MEGABYTE = 1024 * 1024
    _GIGABYTE = 1024 * _MEGABYTE

    MinimumPartSize = 64 * _MEGABYTE
    MaximumNumberOfParts = 10000
    NumberThread = 8
    NumberRetry = 3

    ResponseDataParser = (('ArchiveDescription', 'description', None),
                          ('CreationDate', 'creation_date', None),
                          ('MultipartUploadId', 'id', None),
                          ('PartSizeInBytes', 'part_size', 0))

    def __init__(self, vault, response, file_path=None):
        self.vault = vault
        for response_name, attr_name, default in self.ResponseDataParser:
            value = response.get(response_name)
            setattr(self, attr_name, value or default)
        self.parts = OrderedDict()
        self.file_path = file_path
        self.size_total = 0

        if self.file_path is not None:
            self._prepare(self.file_path)

    def __repr__(self):
        return 'Multipart Upload: %s' % self.id

    @classmethod
    def calc_part_size(cls, size_total, part_size=MinimumPartSize):
        if size_total > 4*10000 * cls._GIGABYTE:
            raise ValueError('File too big: %d' % size_total)

        if size_total < cls.MinimumPartSize:
            raise ValueError('File too small: %d, '
                             'please use vault.upload_archive' %
                             size_total)

        if part_size % cls._MEGABYTE != 0 or \
                part_size < cls.MinimumPartSize or \
                part_size > size_total:
            part_size = cls.MinimumPartSize

        number_parts = calc_num_part(part_size, size_total)
        if number_parts > cls.MaximumNumberOfParts:
            part_size_refer = math.ceil(
                size_total / cls.MaximumNumberOfParts / cls._MEGABYTE)
            part_size_refer = int(part_size_refer * cls._MEGABYTE)
            part_bit_len = part_size_refer.bit_length() - 1
            part_size = 1<< part_bit_len
            while True:
                if(part_size >= part_size_refer):
                    break
                part_bit_len += 1
                part_size = 1<< part_bit_len
        return part_size

    @classmethod
    def parse_range_from_str(cls, string):
        return tuple([int(num) for num in string.split('-')])

    def _prepare(self, file_path):
        self.file_path = file_path
        f = open_file(self.file_path)
        with f:
            self.size_total = content_length(f)
        for byte_range in calc_ranges(self.part_size, self.size_total):
            self.parts[byte_range] = None

    def resume(self, file_path):
        self._prepare(file_path)

        result = self.vault.api.list_parts(self.vault.id, self.id)

        for parts in result['Parts']:
            byte_range = self.parse_range_from_str(parts['RangeInBytes'])
            self.parts[byte_range] = (
                parts['ContentEtag'], parts['TreeEtag'])
            etag = compute_etag_from_file(
                self.file_path, offset=byte_range[0],
                size=range_size(byte_range))
            if etag != self.parts[byte_range][0]:
                raise HashDoesNotMatchError(
                    'Hash does not match for %d-%d: %s, which %s excepted' %
                    (byte_range[0], byte_range[1], etag,
                     self.parts[byte_range][0]))

        return self.start()

    def start(self):
        def upload_part(byte_range):
            try:
                time.sleep(random.randint(256, 4096) / 1000)

                offset = byte_range[0]
                size = range_size(byte_range)
                etag = compute_etag_from_file(
                    self.file_path, offset=offset, size=size)
                tree_etag = compute_tree_etag_from_file(
                    self.file_path, offset=offset, size=size)

                f = open_file(self.file_path)
                with f:
                    for cnt in xrange(self.NumberRetry):
                        try:
                            if self.part_size % mmap.ALLOCATIONGRANULARITY == 0:
                                target = mmap.mmap(f.fileno(), length=size,
                                                   offset=offset,
                                                   access=mmap.ACCESS_READ)
                            else:
                                f.seek(offset)
                                target = f

                            self.vault.api.upload_part(
                                self.vault.id, self.id, target, byte_range,
                                etag = etag, tree_etag=tree_etag)
                            self.parts[byte_range] = (etag, tree_etag)
                            log.info('Range %d-%d upload success.' % byte_range)
                            return
                        except OASServerError as e:
                            log.error('upload %s range %d-%d upload failed. etag: %s. Reason: %s' %
                                      (self.id, byte_range[0], byte_range[1], etag, e))
                            if e.code != 'InvalidDigest' or e.type != 'client':
                                return
                        except IOError as e:
                            log.error('uploadid %s range %d-%d upload failed. Reason: %s' %
                                      (self.id, byte_range[0], byte_range[1], e))
                            continue
                        except Exception as e:
                            log.error('upload %s range %d-%d upload failed. etag: %s. Error: %s' %
                                      (self.id, byte_range[0], byte_range[1], etag, e))
                            continue
                        finally:
                            if 'target' in locals() and target is not f:
                                target.close()
            except Exception as e:
                log.error('Upload %s range %d-%d upload finally failed. Reason: %s' %
                               (self.id, byte_range[0], byte_range[1], e))
                raise


        log.info('Start upload %s from %s.' % (self.id, self.file_path))
        try:
            pool = ThreadPool(processes=min(self.NumberThread, len(self.parts)))
            pool.map(upload_part, [byte_range
                                   for byte_range, tag in self.parts.items()
                                   if tag is None])

            size = self.size_completed
            if size != self.size_total:
                raise UploadArchiveError(
                    'Incomplete upload %s : %d / %d' % (self.id, size, self.size_total))


            response = self.vault.api.complete_multipart_upload(
                self.vault.id, self.id, self.size_total,
                 self.tree_hash)

            log.info('Upload %s finish.' % (self.id ))
            return response['x-oas-archive-id']
        except UploadArchiveError as e:
            errorinfo = 'upload %s failed, cause:%s' % (self.id, e)
            log.error(errorinfo)
            raise e
        except Exception as e:
            errorinfo = 'upload %s failed, cause:%s' % (self.id, e)
            log.error(errorinfo)
            raise ValueError(errorinfo)

    def cancel(self):
        return self.vault.api.cancel_multipart_upload(self.vault.id, self.id)

    @property
    def size_completed(self):
        size_list = [range_size(byte_range)
                     for byte_range, tag in self.parts.items()
                     if tag is not None]
        return sum(size_list)


    @property
    def tree_hash(self):
        tree_etag_list = [tag[1] for _, tag in self.parts.items()
                          if tag is not None]
        return MerkleTree().load(tree_etag_list).digest() \
            if len(tree_etag_list) == len(self.parts) else ''



#!/usr/bin/env python
try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

import os
import sys

def read(*rnames):
    return open(os.path.join(os.path.dirname(__file__), *rnames)).read()

if float('%d.%d' % sys.version_info[:2]) < 2.6 or float('%d.%d' % sys.version_info[:2]) >= 3.0:
    sys.stderr.write("Your Python version %d.%d.%d is not supported.\n" %
                     sys.version_info[:3])
    sys.stderr.write("OAS Python SDK requires Python between 2.6 and 3.0.\n")
    sys.exit(1)

classifiers = [
    'Development Status :: 4 - Beta',

    'Intended Audience :: Developers',
    'Topic :: Software Development :: Libraries :: Python Modules',

    'Operating System :: OS Independent',

    'License :: OSI Approved :: GNU General Public License v2 (GPLv2)',

    'Programming Language :: Python :: 2.6',
    'Programming Language :: Python :: 2.7',
]

setup(
    name='oassdk',
    version='0.2.3',
    description='Python SDK for Aliyun OAS (Open Archive Service)',
    author='Aliyun OAS',
    author_email='jianyi.weng@alibaba-inc.com',
    url='http://www.aliyun.com/product/oas',
    packages=['oas', 'oas.ease'],
    license='GPL version 2',
    install_requires=[
        'pyaml',
        'ordereddict',
    ],
    scripts=['oascmd.py'],
)



# -*- coding: utf-8 -*-
# Copyright (C) 2011  Alibaba Cloud Computing
# 
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.
# 
import io
import json
import yaml


class OASResponse(dict):
    def __init__(self, response):
        super(dict, self).__init__()

        raw_headers = response.getheaders()
        headers = dict()
        for k, v in raw_headers:
            headers[k.lower()] = v

        self.status = response.status
        self.request_id = headers['x-oas-request-id']
        self.response = response

        for k, v in headers.items():
            self[k] = v

        content_type = headers.get('content-type')
        if content_type == 'application/octet-stream':
            self.reader = response
            return

        try:
            body = response.read()
            self.reader = io.BytesIO(body)
            if 'content-range' in self:
                return
            if content_type == 'application/json':
                self.update(json.loads(body))
            else:
                content = yaml.load(body)
                if content is not None:
                    self.update(content)
        except Exception as e:
            raise IOError(e.message)

    def read(self, size):
        return self.reader.read(size)
